/*
 *  arm cpu init and loop
 *
 *  Olivier Houchard
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _TARGET_ARCH_CPU_H_
#define _TARGET_ARCH_CPU_H_

#include "target_arch.h"

/* #define DEBUG_ARM */

#define TARGET_DEFAULT_CPU_MODEL "any"

#define TARGET_CPU_RESET(env)

static inline void target_cpu_init(CPUARMState *env,
        struct target_pt_regs *regs)
{
    int i;

    cpsr_write(env, regs->uregs[16], 0xffffffff);
    for (i = 0; i < 16; i++) {
        env->regs[i] = regs->uregs[i];
    }
}

static inline int do_strex(CPUARMState *env)
{
    uint32_t val;
    int size;
    int rc = 1;
    int segv = 0;
    uint32_t addr;
    start_exclusive();
    addr = env->exclusive_addr;
    if (addr != env->exclusive_test) {
        goto fail;
    }
    size = env->exclusive_info & 0xf;
    switch (size) {
    case 0:
        segv = get_user_u8(val, addr);
        break;
    case 1:
        segv = get_user_u16(val, addr);
        break;
    case 2:
    case 3:
        segv = get_user_u32(val, addr);
        break;
    default:
        abort();
    }
    if (segv) {
        env->cp15.c6_data = addr;
        goto done;
    }
    if (val != env->exclusive_val) {
        goto fail;
    }
    if (size == 3) {
        segv = get_user_u32(val, addr + 4);
        if (segv) {
            env->cp15.c6_data = addr + 4;
            goto done;
        }
        if (val != env->exclusive_high) {
            goto fail;
        }
    }
    val = env->regs[(env->exclusive_info >> 8) & 0xf];
    switch (size) {
    case 0:
        segv = put_user_u8(val, addr);
        break;
    case 1:
        segv = put_user_u16(val, addr);
        break;
    case 2:
    case 3:
        segv = put_user_u32(val, addr);
        break;
    }
    if (segv) {
        env->cp15.c6_data = addr;
        goto done;
    }
    if (size == 3) {
        val = env->regs[(env->exclusive_info >> 12) & 0xf];
        segv = put_user_u32(val, addr + 4);
        if (segv) {
            env->cp15.c6_data = addr + 4;
            goto done;
        }
    }
    rc = 0;
fail:
    env->regs[15] += 4;
    env->regs[(env->exclusive_info >> 4) & 0xf] = rc;
done:
    end_exclusive();
    return segv;
}

static inline void target_cpu_loop(CPUARMState *env)
{
    int trapnr;
    unsigned int n, insn;
    uint32_t addr;
    CPUState *cs = CPU(arm_env_get_cpu(env));

    for (;;) {
#ifdef DEBUG_ARM
        printf("CPU LOOPING\n");
#endif
        cpu_exec_start(cs);
#ifdef DEBUG_ARM
        printf("EXECUTING...\n");
#endif
        trapnr = cpu_arm_exec(env);
#ifdef DEBUG_ARM
        printf("trapnr %d\n", trapnr);
#endif
        cpu_exec_end(cs);
        switch (trapnr) {
        case EXCP_UDEF:
            {
#if 0
                TaskState *ts = env->opaque;
                uint32_t opcode;
                int rc;

                /* we handle the FPU emulation here, as Linux */
                /* we get the opcode */
                /* FIXME - what to do if get_user() fails? */
                get_user_u32(opcode, env->regs[15]);

                rc = EmulateAll(opcode, &ts->fpa, env);
                if (rc == 0) { /* illegal instruction */
                    info.si_signo = SIGILL;
                    info.si_errno = 0;
                    info.si_code = TARGET_ILL_ILLOPN;
                    info._sifields._sigfault._addr = env->regs[15];
                    queue_signal(env, info.si_signo, &info);

                } else if (rc < 0) { /* FP exception */
                    int arm_fpe = 0;

                    /* translate softfloat flags to FPSR flags */
                    if (-rc & float_flag_invalid) {
                        arm_fpe |= BIT_IOC;
                    }
                    if (-rc & float_flag_divbyzero) {
                        arm_fpe |= BIT_DZC;
                    }
                    if (-rc & float_flag_overflow) {
                        arm_fpe |= BIT_OFC;
                    }
                    if (-rc & float_flag_underflow) {
                        arm_fpe |= BIT_UFC;
                    }
                    if (-rc & float_flag_inexact) {
                        arm_fpe |= BIT_IXC;
                    }

                    FPSR fpsr = ts->fpa.fpsr;
                    /* printf("fpsr 0x%x, arm_fpe 0x%x\n",fpsr,arm_fpe); */

                    if (fpsr & (arm_fpe << 16)) { /* exception enabled? */
                        info.si_signo = SIGFPE;
                        info.si_errno = 0;

                        /* ordered by priority, least first */
                        if (arm_fpe & BIT_IXC) {
                            info.si_code = TARGET_FPE_FLTRES;
                        }
                        if (arm_fpe & BIT_UFC) {
                            info.si_code = TARGET_FPE_FLTUND;
                        }
                        if (arm_fpe & BIT_OFC) {
                            info.si_code = TARGET_FPE_FLTOVF;
                        }
                        if (arm_fpe & BIT_DZC) {
                            info.si_code = TARGET_FPE_FLTDIV;
                        }
                        if (arm_fpe & BIT_IOC) {
                            info.si_code = TARGET_FPE_FLTINV;
                        }
                        info._sifields._sigfault._addr = env->regs[15];
                        queue_signal(env, info.si_signo, &info);
                    } else {
                        env->regs[15] += 4;
                    }

                    /* accumulate unenabled exceptions */
                    if ((!(fpsr & BIT_IXE)) && (arm_fpe & BIT_IXC)) {
                        fpsr |= BIT_IXC;
                    }
                    if ((!(fpsr & BIT_UFE)) && (arm_fpe & BIT_UFC)) {
                        fpsr |= BIT_UFC;
                    }
                    if ((!(fpsr & BIT_OFE)) && (arm_fpe & BIT_OFC)) {
                        fpsr |= BIT_OFC;
                    }
                    if ((!(fpsr & BIT_DZE)) && (arm_fpe & BIT_DZC)) {
                        fpsr |= BIT_DZC;
                    }
                    if ((!(fpsr & BIT_IOE)) && (arm_fpe & BIT_IOC)) {
                        fpsr |= BIT_IOC;
                    }
                    ts->fpa.fpsr = fpsr;
                } else { /* everything OK */
                    /* increment PC */
                    env->regs[15] += 4;
                }
            }
#endif
            break;
        case EXCP_SWI:
        case EXCP_BKPT:
            {
                env->eabi = 1;
                /* system call */
                if (trapnr == EXCP_BKPT) {
                    if (env->thumb) {
                        /* FIXME - what to do if get_user() fails? */
                        get_user_u16(insn, env->regs[15]);
                        n = insn & 0xff;
                        env->regs[15] += 2;
                    } else {
                        /* FIXME - what to do if get_user() fails? */
                        get_user_u32(insn, env->regs[15]);
                        n = (insn & 0xf) | ((insn >> 4) & 0xff0);
                        env->regs[15] += 4;
                    }
                } else {
                    if (env->thumb) {
                        /* FIXME - what to do if get_user() fails? */
                        get_user_u16(insn, env->regs[15] - 2);
                        n = insn & 0xff;
                    } else {
                        /* FIXME - what to do if get_user() fails? */
                        get_user_u32(insn, env->regs[15] - 4);
                        n = insn & 0xffffff;
                    }
                }

#ifdef DEBUG_ARM
        printf("AVANT CALL %d\n", n);
#endif
                if (bsd_type == target_freebsd) {
                    int ret;
                    abi_ulong params = get_sp_from_cpustate(env);
                    int32_t syscall_nr = n;
                    int32_t arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8;

#if 0 /* XXX FIXME */
                    if (syscall_nr == TARGET_FREEBSD_NR_syscall) {
                        get_user_s32(syscall_nr, params);
                        params += sizeof(int32_t);
                    } else if (syscall_nr == TARGET_FREEBSD_NR___syscall) {
                        get_user_s32(syscall_nr, params);
                        params += sizeof(int64_t);
                    }
#endif
                    arg1 = env->regs[0];
                    arg2 = env->regs[1];
                    arg3 = env->regs[2];
                    arg4 = env->regs[3];
                    get_user_s32(arg5, params);
                    params += sizeof(int32_t);
                    get_user_s32(arg6, params);
                    params += sizeof(int32_t);
                    get_user_s32(arg7, params);
                    params += sizeof(int32_t);
                    get_user_s32(arg8, params);
                    ret = do_freebsd_syscall(env,
                                                      syscall_nr,
                                                      arg1,
                                                      arg2,
                                                      arg3,
                                                      arg4,
                                                      arg5,
                                                      arg6,
                                                      arg7,
                                                      arg8);
                    /*
                     * Compare to arm/arm/vm_machdep.c
                     * cpu_set_syscall_retval()
                     */
                    /* XXX armeb may need some extra magic here */
                    if (-TARGET_EJUSTRETURN == ret) {
                        /*
                         * Returning from a successful sigreturn syscall.
                         * Avoid clobbering register state.
                         */
                        break;
                    }
                    /*
                     * XXX Need to handle ERESTART. Backup the PC by
                     * 1 instruction.
                     */
                    if ((unsigned int)ret >= (unsigned int)(-515)) {
                        ret = -ret;
                        cpsr_write(env, CPSR_C, CPSR_C);
                        env->regs[0] = ret;
                    } else {
                        cpsr_write(env, 0, CPSR_C);
                        env->regs[0] = ret; /* XXX need to handle lseek()? */
                        /* env->regs[1] = 0; */
                    }
                } else {
                    /* XXX is this correct? */
                    env->regs[0] = do_openbsd_syscall(env,
                        n,
                        env->regs[0],
                        env->regs[1],
                        env->regs[2],
                        env->regs[3],
                        env->regs[4],
                        env->regs[5]);
                }
#ifdef DEBUG_ARM
                printf("APRES CALL\n");
#endif
            }
        }
            break;
        case EXCP_INTERRUPT:
            /* just indicate that signals should be handled asap */
            break;
        case EXCP_PREFETCH_ABORT:
            addr = env->cp15.c6_insn;
            goto do_segv;
        case EXCP_DATA_ABORT:
            addr = env->cp15.c6_data;
        do_segv:
            {
#if 0
                info.si_signo = SIGSEGV;
                info.si_errno = 0;
                /* XXX: check env->error_code */
                info.si_code = TARGET_SEGV_MAPERR;
                info._sifields._sigfault._addr = addr;
                queue_signal(env, info.si_signo, &info);
#endif
            }
            break;
        case EXCP_DEBUG:
            {
                int sig;

                sig = gdb_handlesig(env, TARGET_SIGTRAP);
                if (sig) {
#if 0
                    info.si_signo = sig;
                    info.si_errno = 0;
                    info.si_code = TARGET_TRAP_BRKPT;
                    queue_signal(env, info.si_signo, &info);
#endif
                }
            }
            break;
#if 0
        case EXCP_KERNEL_TRAP:
            if (do_kernel_trap(env)) {
                goto error;
            }
            break;
#endif
        case EXCP_STREX:
            if (do_strex(env)) {
                addr = env->cp15.c6_data;
                goto do_segv;
            }
            break;
#if 0
        error:
#endif
        default:
            fprintf(stderr, "qemu: unhandled CPU exception 0x%x - aborting\n",
                    trapnr);
            cpu_dump_state(env, stderr, fprintf, 0);
            abort();
        }
        process_pending_signals(env);
    }
}

#endif /* !_TARGET_ARCH_CPU_H */
