#ifndef _TARGET_OS_STACK_H_
#define _TARGET_OS_STACK_H_

#include <sys/param.h>
#include "target_arch_sigtramp.h"

/*
 * The inital FreeBSD stack is as follows:
 * (see kern/kern_exec.c exec_copyout_strings() )
 *
 *  Hi Address -> char **ps_argvstr  (struct ps_strings for ps, w, etc.)
 *                unsigned ps_nargvstr
 *                char **ps_envstr
 *  PS_STRINGS -> unsigned ps_nenvstr
 *
 *                machine dependent sigcode (sv_sigcode of size
 *                                           sv_szsigcode)
 *
 *                execpath          (absolute image path for rtld)
 *
 *                SSP Canary        (sizeof(long) * 8)
 *
 *                page sizes array  (usually sizeof(u_long) )
 *
 *  "destp" ->    argv, env strings (up to 262144 bytes)
 */
static inline abi_ulong setup_initial_stack(abi_ulong p,
        struct bsd_binprm *bprm, abi_ulong stack_base, abi_ulong size,
        abi_ulong addr)
{
    int i;
    abi_ulong stack_hi_addr;
    size_t execpath_len;
    abi_ulong destp, argvp, envp;
    struct target_ps_strings ps_strs;
    char canary[sizeof(abi_long) * 8];

    stack_hi_addr = p = addr + size;

    /* Save some space for ps_strings. */
    p -= sizeof(struct target_ps_strings);

#ifdef TARGET_SZSIGCODE
    /* Add machine depedent sigcode. */
    p -= TARGET_SZSIGCODE;
    if (setup_sigtramp(p, (unsigned)offsetof(struct target_sigframe, sf_uc),
            TARGET_FREEBSD_NR_sigreturn)) {
        return 0;
    }
#endif
    if (bprm->fullpath) {
        execpath_len = strlen(bprm->fullpath) + 1;
        p -= roundup(execpath_len, sizeof(abi_ulong));
        if (memcpy_to_target(p, bprm->fullpath, execpath_len)) {
            return 0;
        }
    }
    /* Add canary for SSP. */
    arc4random_buf(canary, sizeof(canary));
    p -= roundup(sizeof(canary), sizeof(abi_ulong));
    if (memcpy_to_target(p, canary, sizeof(canary))) {
        return 0;
    }
    /* Add page sizes array. */
    p -= sizeof(abi_ulong);
    if (put_user_ual(TARGET_PAGE_SIZE, p)) {
        return 0;
    }
    argvp = p - TARGET_SPACE_USRSPACE;
    p = destp = p - TARGET_SPACE_USRSPACE - TARGET_ARG_MAX;
    /* XXX check strlen(argv and envp strings) < TARGET_ARG_MAX ? */

    /*
     * Add argv strings.  Note that the argv[] vectors are added by
     * loader_build_argptr()
     */
    envp = argvp + (bprm->argc + 2) * sizeof(abi_ulong);
    ps_strs.ps_argvstr = tswapl(argvp);
    ps_strs.ps_nargvstr = tswap32(bprm->argc);
    for (i = 0; i < bprm->argc; ++i) {
        size_t len = strlen(bprm->argv[i]) + 1;

        if (memcpy_to_target(destp, bprm->argv[i], len)) {
            return 0;
        }
        if (put_user_ual(destp, argvp)) {
            return 0;
        }
        argvp += sizeof(abi_ulong);
        destp += len;
    }
    if (put_user_ual(0, argvp)) {
        return 0;
    }
    /*
     * Add env strings. Note that the envp[] vectors are added by
     * loader_build_argptr().
     */
    ps_strs.ps_envstr = tswapl(envp);
    ps_strs.ps_nenvstr = tswap32(bprm->envc);
    for (i = 0; i < bprm->envc; ++i) {
        size_t len = strlen(bprm->envp[i]) + 1;

        if (memcpy_to_target(destp, bprm->envp[i], len)) {
            return 0;
        }
        if (put_user_ual(destp, envp)) {
            return 0;
        }
        envp += sizeof(abi_ulong);
        destp += len;
    }
    if (put_user_ual(0, envp)) {
        return 0;
    }
    if (memcpy_to_target(stack_hi_addr - sizeof(ps_strs), &ps_strs,
                sizeof(ps_strs))) {
        return 0;
    }

    return p;
 }

#endif /* !_TARGET_OS_STACK_H_ */
